/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIAYSTOK 2014                               **
**                                                                        **
**                         Home Control System                            **
**                                                                        **
**                              HC_PWC2                                   **
**                      Wersja z dnia : 2017.12.09                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include "pwc2eeprom.h"

#define _KeySmall_

#include "..\common\hc_keycodes.h"

#define HardVer200


#define false                          0
#define true                           1

#define PowerOnOffLed                  1
#define StatusLed                      0
#define RxTxLed                        2
#define KeyLed                         3

#define KeybPort                       PORTD
#define KeybPortDirection              DDRD
#define InpKeybReg                     PIND

#define LKey                           5              // PORTD 5

#define TimerCounterDelayLimit         3000  // 2048
#define StandardLEDPulseDelay          832
#define TimerPrescaller                3600

#define LEDPort                        PORTC
#define LEDPortDirection               DDRC

#define PowerPort                      PORTB
#define PowerPortDirection             DDRB

#define PwrSwitchPin                   0

#define RxDPin                         0              // PORTD . 0
#define TxDPin                         1              // PORTD . 1
#define RxTxDirPin                     2              // PORTD . 2

#define null                           0
#define cr                             0x0D
#define Space                          0x20
#define PartSeparator                  ':'
#define ParamSeparator                 '.'

#define nop() __asm__ __volatile__ ("nop")

// USART BAUD Rate Generator fosc=7.3728 MHz
#define Speed9600bps                   47
#define Speed19200bps                  23
#define Speed38400bps                  11
#define Speed115200bps                 3

#define IdleState                      0
#define BusyState                      1

#define TXDBuffSize                    128
#define RXDBuffSize                    32

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t State ;
                 uint8_t Buffer [ TXDBuffSize ] ;
               } TXDRecT ;
typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ RXDBuffSize ] ;
               } RXDRecT ;

#define CyclBuffSize                   256

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ CyclBuffSize ] ;
               } CyclicRecT ;

typedef void ( * OperationProcT ) ( void ) ;

typedef struct {
                 volatile TXDRecT          TXDRec ;
                 volatile RXDRecT          RXDRec ;
                 OperationProcT            SerialService ;
                 CommandBufferType         CommandText ;
                 ParamBufferType           ParamText ;
                 DeviceIDType              DestinationDeviceName ;
                 DeviceIDType              SourceDeviceName ;
                 uint8_t                   CommandTextIndex ;
                 uint8_t                   ParamTextIndex ;
                 uint8_t                   DestDeviceNameIndex ;
                 uint8_t                   SrcDeviceNameIndex ;
                 uint16_t                  SerialLEDDelay ;
               } ServiceType ;

typedef ServiceType * PtrServiceType ;

#define TempEchoDevicesArraySize       4

typedef enum { KeybIdle ,
               KeybWaitOnPress ,
               KeybReady ,
               KeybActive ,
               KeybWaitOnFree ,
               KeybWait
             } KeybStatusType ;

#define KeybDelayLimit                 2
#define KeybLongDelayLimit             25

static ServiceType SerialService ;
static KeybStatusType KeybStatus ;
static uint8_t LastKeybInp ;
static uint8_t KeybCode ;
static uint16_t KeybDelayCt ;
static uint16_t KeybDelayTimer ;
static uint16_t AnyKeyDelay ;
static uint8_t ConnectContext ;
static volatile uint8_t TimeEvent ;
static uint16_t TimerCounter ;
static uint8_t CurrentChannelIsOn ;
static DeviceIDType TempEchoDevices [ TempEchoDevicesArraySize ] ;
static CyclicRecT CyclicBuffer ;
static uint32_t ServiceDelayTime ;
static OperationProcT DelayedService ;
static uint8_t DelayedServiceTimeEvent ;
static uint8_t KeyCommandParam ;
static uint16_t BlockEECurrentAddress ;


static const uint16_t CRCTable [ 16 ] = { 0x0000 ,  /* 00 */
                                          0xCC01 ,  /* 01 */
                                          0xD801 ,  /* 02 */
                                          0x1400 ,  /* 03 */
                                          0xF001 ,  /* 04 */
                                          0x3C00 ,  /* 05 */
                                          0x2800 ,  /* 06 */
                                          0xE401 ,  /* 07 */
                                          0xA001 ,  /* 08 */
                                          0x6C00 ,  /* 09 */
                                          0x7800 ,  /* 10 */
                                          0xB401 ,  /* 11 */
                                          0x5000 ,  /* 12 */
                                          0x9C01 ,  /* 13 */
                                          0x8801 ,  /* 14 */
                                          0x4400 } ;/* 15 */

static uint8_t HardwareSoftwareVersion [ ] PROGMEM   = "HVER*2_00*V-2013.SVER*1_3*09-12-2017\r" ;
static uint8_t DeviceName [ ] PROGMEM                = "HCPWC2-power controller\r" ;
static uint8_t MyDefaultID [ ] PROGMEM               = "PWC2" ;
static uint8_t GeneralPoolDeviceName [ ] PROGMEM     = "GENERAL" ;
static uint8_t OK [ ] PROGMEM                        = "<OK>\r" ;
static uint8_t Fail [ ] PROGMEM                      = "<FA>\r" ;
static uint8_t HexalTable [ ] PROGMEM                = "0123456789ABCDEF" ;
static uint8_t NoDataText [ ] PROGMEM                = "NODATA\r" ;
static uint8_t EmptyFlagTxt [ ] PROGMEM              = "<EMPTY>\r" ;

static uint8_t ChannelIsOnState [ ] PROGMEM          = "CHISON\r" ;
static uint8_t ChannelIsOffState [ ] PROGMEM         = "CHISOF\r" ;

static uint8_t ReplyCommand [ ] PROGMEM              = "REPLY" ;
static uint8_t ConnectCommand [ ] PROGMEM            = "CONNECT" ;
static uint8_t DisConnectCommand [ ] PROGMEM         = "DISCONNECT" ;
static uint8_t SetDeviceNameCommand [ ] PROGMEM      = "SETMID" ;
static uint8_t HelloCommand [ ] PROGMEM              = "HELLO" ;
static uint8_t GetVersionCommand [ ] PROGMEM         = "GETVER" ;
static uint8_t GetIDCommand [ ] PROGMEM              = "GETID" ;
static uint8_t ReinitCommand [ ] PROGMEM             = "REINIT" ;
static uint8_t ResetEECommand [ ] PROGMEM            = "RESEE" ;
static uint8_t SaveEECommand [ ] PROGMEM             = "SAVEEE" ;
static uint8_t GetChnStatusCommand [ ] PROGMEM       = "GETCHNST" ;
static uint8_t AddSuperiorChnCommand [ ] PROGMEM     = "ADDSUP" ;
static uint8_t DelSuperiorChnCommand [ ] PROGMEM     = "DELSUP" ;
static uint8_t ListSuperiorChnCommand [ ] PROGMEM    = "LISTSUP" ;
static uint8_t ChStateServiceOnCommand [ ] PROGMEM   = "CSSERVON" ;
static uint8_t ChStateServiceOffCommand [ ] PROGMEM  = "CSSERVOFF" ;
static uint8_t ChStateServiceListCommand [ ] PROGMEM = "CSSERVLIST" ;
static uint8_t ListKeyTableCommand [ ] PROGMEM       = "LISTKC" ;
static uint8_t AddKeyTableCommand [ ] PROGMEM        = "ADDKC" ;
static uint8_t DelKeyTableCommand [ ] PROGMEM        = "DELKC" ;
static uint8_t NewPasswordCommand [ ] PROGMEM        = "NEWPW" ;
static uint8_t RunKeyCommand [ ] PROGMEM             = "RUNKC" ;

typedef void ( * CommandExecuteServiceProcT ) ( void ) ;
typedef struct {
                 CommandExecuteServiceProcT  CommandService ;
                 uint8_t *                   CommandName ;
               } CommandNameAssociationRecT ;

static void ReplyCommandExecute ( void ) ;
static void ConnectCommandExecute ( void ) ;
static void DisConnectCommandExecute ( void ) ;
static void SetDeviceNameCommandExecute ( void ) ;
static void HelloCommandExecute ( void ) ;
static void GetVersionCommandExecute ( void ) ;
static void GetIDCommandExecute ( void ) ;
static void ReinitCommandExecute ( void ) ;
static void ResetEECommandExecute ( void ) ;
static void SaveEECommandExecute ( void ) ;
static void GetChnStatusCommandExecute ( void ) ;
static void AddSuperiorChnCommandExecute ( void ) ;
static void DelSuperiorChnCommandExecute ( void ) ;
static void ChStateServiceOnCommandExecute ( void ) ;
static void ChStateServiceOffCommandExecute ( void ) ;
static void ChStateServiceListCommandExecute ( void ) ;
static void ListKeyTableCommandExecute ( void ) ;
static void AddKeyTableCommandExecute ( void ) ;
static void DelKeyTableCommandExecute ( void ) ;
static void ListSuperiorChnCommandExecute ( void ) ;
static void NewPasswordCommandExecute ( void ) ;
static void RunKeyCommandExecute ( void ) ;
static void ChOnCommandExecute ( void ) ;
static void ChOffCommandExecute ( void ) ;
static void ChInvCommandExecute ( void ) ;
static void ChPulseOnCommandExecute ( void ) ;
static void ChPulseOffCommandExecute ( void ) ;
static void ChLongPulseOnCommandExecute ( void ) ;
static void ChLongPulseOffCommandExecute ( void ) ;


#define CommandNameAssociationTabeSize               29

static CommandNameAssociationRecT CommandNameAssociationTabe [ CommandNameAssociationTabeSize ] PROGMEM =
{
/* 00 */ { ChOnCommandExecute                  , ChOnCommand } ,
/* 01 */ { ChOffCommandExecute                 , ChOffCommand } ,
/* 02 */ { ChInvCommandExecute                 , ChInvCommand } ,
/* 03 */ { ChPulseOnCommandExecute             , ChPulseOnCommand } ,
/* 04 */ { ChPulseOffCommandExecute            , ChPulseOffCommand } ,
/* 05 */ { ChLongPulseOnCommandExecute         , ChLongPulseOnCommand } ,
/* 06 */ { ChLongPulseOffCommandExecute        , ChLongPulseOffCommand } ,
/* 07 */ { ReplyCommandExecute                 , ReplyCommand } ,
/* 08 */ { ConnectCommandExecute               , ConnectCommand } ,
/* 09 */ { DisConnectCommandExecute            , DisConnectCommand } ,
/* 10 */ { SetDeviceNameCommandExecute         , SetDeviceNameCommand } ,
/* 11 */ { HelloCommandExecute                 , HelloCommand } ,
/* 12 */ { GetVersionCommandExecute            , GetVersionCommand } ,
/* 13 */ { GetIDCommandExecute                 , GetIDCommand } ,
/* 14 */ { ReinitCommandExecute                , ReinitCommand } ,
/* 15 */ { ResetEECommandExecute               , ResetEECommand } ,
/* 16 */ { SaveEECommandExecute                , SaveEECommand } ,
/* 17 */ { GetChnStatusCommandExecute          , GetChnStatusCommand } ,
/* 18 */ { AddSuperiorChnCommandExecute        , AddSuperiorChnCommand } ,
/* 19 */ { DelSuperiorChnCommandExecute        , DelSuperiorChnCommand } ,
/* 20 */ { ListSuperiorChnCommandExecute       , ListSuperiorChnCommand } ,
/* 21 */ { ChStateServiceOnCommandExecute      , ChStateServiceOnCommand } ,
/* 22 */ { ChStateServiceOffCommandExecute     , ChStateServiceOffCommand } ,
/* 23 */ { ChStateServiceListCommandExecute    , ChStateServiceListCommand } ,
/* 24 */ { ListKeyTableCommandExecute          , ListKeyTableCommand } ,
/* 25 */ { AddKeyTableCommandExecute           , AddKeyTableCommand } ,
/* 26 */ { DelKeyTableCommandExecute           , DelKeyTableCommand } ,
/* 27 */ { NewPasswordCommandExecute           , NewPasswordCommand } ,
/* 28 */ { RunKeyCommandExecute                , RunKeyCommand } ,
} ;

static void ChOnKeyCommandExecute ( void ) ;
static void ChOffKeyCommandExecute ( void ) ;
static void ChInvKeyCommandExecute ( void ) ;
static void ChPulseOnKeyCommandExecute ( void ) ;
static void ChPulseOffKeyCommandExecute ( void ) ;
static void ChLongPulseOnKeyCommandExecute ( void ) ;
static void ChLongPulseOffKeyCommandExecute ( void ) ;

#define KeyCommandAssociationTabeSize                7

static CommandNameAssociationRecT KeyCommandAssociationTabe [ KeyCommandAssociationTabeSize ] PROGMEM =
{
/* 00 */ { ChOnKeyCommandExecute               , ChOnCommand } ,
/* 01 */ { ChOffKeyCommandExecute              , ChOffCommand } ,
/* 02 */ { ChInvKeyCommandExecute              , ChInvCommand } ,
/* 03 */ { ChPulseOnKeyCommandExecute          , ChPulseOnCommand } ,
/* 04 */ { ChPulseOffKeyCommandExecute         , ChPulseOffCommand } ,
/* 05 */ { ChLongPulseOnKeyCommandExecute      , ChLongPulseOnCommand } ,
/* 06 */ { ChLongPulseOffKeyCommandExecute     , ChLongPulseOffCommand } ,
} ;




static uint8_t KeyPressed ( void )
{
  /*----------------------------------------------------------------*/
  return ( KeybCode != 0 ) ;
} /* KeyPressed */


static uint8_t ReadKey ( void )
{
  uint8_t TempCh ;
  /*----------------------------------------------------------------*/
  TempCh = KeybCode ;
  KeybCode = 0 ;
  return ( TempCh ) ;
} /* ReadKey */


static uint8_t HardwareKeybRead ( void )
{
  uint8_t KeybData ;
  /*----------------------------------------------------------------*/
  KeybData = InpKeybReg ;
  return ( KeybData & ( 1 << LKey ) ) ;
} /* HardwareKeybRead */


static uint8_t AnyKeyActive ( void )
{
  uint8_t KeybData ;
  /*----------------------------------------------------------------*/
  KeybData = HardwareKeybRead ( ) ;
  if ( KeybData )
    return ( false ) ;
  else
    return ( true ) ;
} /* AnyKeyActive */


static void KeybService ( void )
{
  uint8_t KeybActiveStatus ;
  /*----------------------------------------------------------------*/
  KeybActiveStatus = AnyKeyActive ( ) ;
  if ( KeybActiveStatus )
    LEDPort &= ~ ( 1 << KeyLed ) ;
  else
    LEDPort |= 1 << KeyLed ;
  switch ( KeybStatus )
  {
    case KeybIdle        :
      if ( KeybActiveStatus )
      {
        KeybStatus = KeybWaitOnPress ;
        KeybDelayCt = KeybDelayLimit ;
      } /* if */ ;
      break ;
    case KeybWaitOnPress :
      KeybDelayCt -- ;
      if ( ! KeybDelayCt )
      {
        if ( KeybActiveStatus )
          KeybStatus = KeybReady ;
        else
          KeybStatus = KeybIdle ;
      } /* if */ ;
      break ;
    case KeybReady       :
      LastKeybInp = ( HardwareKeybRead ( ) ^ 0xFF ) & ( 1 << LKey ) ;
      KeybDelayCt = KeybLongDelayLimit ;
      KeybStatus = KeybActive ;
      AnyKeyDelay = StandardLEDPulseDelay ;
      break ;
    case KeybActive      :
      KeybDelayCt -- ;
      if ( ! KeybDelayCt )
      {
        KeybCode = LongKey0Code ;
        KeybStatus = KeybWaitOnFree ;
        KeybDelayCt = KeybDelayLimit ;
      } /* if ... */
      else
      {
        if ( ! KeybActiveStatus )
        {
          KeybCode = StandKey0Code ;
          KeybStatus = KeybWaitOnFree ;
          KeybDelayCt = KeybDelayLimit ;
        } /* if */ ;
      } /* if ... else */ ;
      break ;
    case KeybWaitOnFree  :
      KeybDelayCt -- ;
      if ( ! KeybDelayCt )
      {
        KeybStatus = KeybWait ;
      } /* if */ ;
      break ;
    case KeybWait        :
      if ( ! KeybActiveStatus )
      {
        KeybStatus = KeybIdle ;
      } /* if */ ;
      break ;
  } /* switch */ ;
} /* KeybService */


static void USARTReceiveMode ( void )
{
  /*----------------------------------------------------------------*/
  PORTD &= ~ ( 1 << RxTxDirPin ) ;
} /* USARTReceiveMode */


static void USARTTransmitMode ( void )
{
  /*----------------------------------------------------------------*/
  PORTD |= ( 1 << RxTxDirPin ) ;
} /* USARTTransmitMode */


SIGNAL ( TIMER0_OVF_vect )
{
  /*----------------------------------------------------------------*/
  TimerCounter ++ ;
  if ( TimerCounter > TimerCounterDelayLimit )
  {
    TimerCounter = 0 ;
    TimeEvent = true ;
  } /* if */ ;
  if ( SerialService . SerialLEDDelay )
  {
    SerialService . SerialLEDDelay -- ;
    if ( ! SerialService . SerialLEDDelay )
    {
      LEDPort |= ( 1 << RxTxLed ) ;
    } /* if */ ;
  } /* if */ ;
  KeybDelayTimer ++ ;
  if ( KeybDelayTimer > 100 )
  {
    KeybDelayTimer = 0 ;
    KeybService ( ) ;
  } /* if */ ;
  if ( ServiceDelayTime )
  {
    ServiceDelayTime -- ;
    if ( ! ServiceDelayTime )
    {
      DelayedServiceTimeEvent = true ;
    } /* if */ ;
  } /* if */ ;
} /* TIMER0_OVF_vect */


static void StartRxTxLED ( void )
{
  /*----------------------------------------------------------------*/
  LEDPort &= ~ ( 1 << RxTxLed ) ;
  SerialService . SerialLEDDelay = StandardLEDPulseDelay ;
} /* StartRxTxLED */


SIGNAL ( USART_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR0 ;
  SerialService . RXDRec . Buffer [ SerialService . RXDRec . CycTo ] = RecCh ;
  SerialService . RXDRec . CycTo ++ ;
  if ( SerialService . RXDRec . CycTo >= RXDBuffSize )
    SerialService . RXDRec . CycTo = 0 ;
  StartRxTxLED ( ) ;
} /* USART_RX_vect */


SIGNAL ( USART_TX_vect )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . TXDRec . CycFr == SerialService . TXDRec . CycTo )
  {
    SerialService . TXDRec . State = IdleState ;
    USARTReceiveMode ( ) ;
  } /* if ... */
  else
  {
    UDR0 = SerialService . TXDRec . Buffer [ SerialService . TXDRec . CycFr ] ;
    SerialService . TXDRec . CycFr ++ ;
    if ( SerialService . TXDRec . CycFr >= TXDBuffSize )
      SerialService . TXDRec . CycFr = 0 ;
  } /* if ... else */ ;
} /* USART_TX_vect */


static void SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  USARTTransmitMode ( ) ;
  StartRxTxLED ( ) ;
  cli ( ) ;
  if ( SerialService . TXDRec . State == IdleState )
  {
    SerialService . TXDRec . State = BusyState ;
    UDR0 = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = SerialService . TXDRec . CycTo + 1 ;
      if ( NextInx >= TXDBuffSize )
        NextInx = 0 ;
      if ( NextInx != SerialService . TXDRec . CycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    SerialService . TXDRec . Buffer [ SerialService . TXDRec . CycTo ] = SerialCh ;
    SerialService . TXDRec . CycTo ++ ;
    if ( SerialService . TXDRec . CycTo >= TXDBuffSize )
      SerialService . TXDRec . CycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* SendSerial */


static void SendSerialTxt ( uint8_t * String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      SendSerial ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* SendSerialTxt */


static void SendSerialFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      SendSerial ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* SendSerialFlashTxt */


static uint8_t SerialDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . RXDRec . CycFr == SerialService . RXDRec . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* SerialDataPresent */


static uint8_t GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = SerialService . RXDRec . Buffer [ SerialService . RXDRec . CycFr ] ;
  SerialService . RXDRec . CycFr ++ ;
  if ( SerialService . RXDRec . CycFr >= RXDBuffSize )
    SerialService . RXDRec . CycFr = 0 ;
  return ( Ch ) ;
} /* GetSerial */


static void AddToCyclic ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  CyclicBuffer . Buffer [ CyclicBuffer . CycTo ] = Ch ;
  CyclicBuffer . CycTo ++ ;
//  if ( CyclicBuffer . CycTo >= CyclBuffSize )
//    CyclicBuffer. CycTo = 0 ;
} /* AddToCyclic */


static void AddToCyclicTxt ( uint8_t * String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* AddToCyclicTxt */


static void AddToCyclicFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* AddToCyclicFlashTxt */


static uint8_t CyclicDataPresent ( void )
{
  /*----------------------------------------------------------------*/
  if ( CyclicBuffer . CycFr == CyclicBuffer . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* CyclicDataPresent */


static uint8_t GetFromCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = CyclicBuffer . Buffer [ CyclicBuffer . CycFr ] ;
  CyclicBuffer . CycFr ++ ;
//  if ( CyclicBuffer . CycFr >= RXDBuffSize )
//    CyclicBuffer . CycFr = 0 ;
  return ( Ch ) ;
} /* GetFromCyclic */


static void FlushCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  while ( CyclicDataPresent ( ) )
  {
    Ch = GetFromCyclic ( );
    SendSerial ( Ch ) ;
  } /* while */ ;
} /* FlushCyclic */


static void USARTHardwareInit ( uint16_t Speed )
{
  /*----------------------------------------------------------------*/
  UCSR0A = 0 ;
  UCSR0B = ( 1 << RXCIE0 ) | ( 1 << TXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ) ;
  UCSR0C = ( 1 << UCSZ01 ) | ( 1 << UCSZ00 ) ;
  UBRR0 = Speed ;
} /* USARTHardwareInit */


static void HardwareInit ( uint16_t Speed )
{
  /*----------------------------------------------------------------*/
  TCCR0A = 0 ;
  TCCR0B = ( 1 << CS01 ) ; // CS02=0 CS01=1 CS00=0 : F clk div 8
  TCNT0 = 0 ;
  OCR0A = 0 ;
  OCR0B = 0 ;
  TIMSK0 = ( 1 << TOIE0 ) ;
  DDRD &= ~ ( 1 << RxDPin ) ;
  DDRD |= ( 1 << TxDPin ) | ( 1 << RxTxDirPin ) ;
  USARTHardwareInit ( Speed ) ;
} /* HardwareInit */


static void EnvirInit ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << PowerOnOffLed ) | ( 1 << StatusLed ) |
         ( 1 << RxTxLed ) | ( 1 << KeyLed ) ;
  LEDPortDirection |= Bits ;
  LEDPort |= Bits ;
  Bits = ( 1 << PwrSwitchPin ) ;
  PowerPortDirection |= Bits ;
  PowerPort |= Bits ;
} /* EnvirInit */


static void ClearBlock ( void *   Destination ,
                         uint16_t Size )
{
  uint16_t Loop ;
  uint8_t * Dst ;
  /*----------------------------------------------------------------*/
  Dst = ( uint8_t * ) Destination ;
  for ( Loop = 0 ; Loop < Size ; Loop ++ )
    * Dst ++ = null ;
} /* ClearBlock */


static void StrCopy ( uint8_t * Destination ,
                      uint8_t * Source )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = * Source ;
  * Destination = Ch ;
  while ( Ch )
  {
    Ch = * Source ++ ;
    * Destination ++ = Ch ;
  } /* while */ ;
} /* StrCopy */


static void Assign ( uint8_t * SubDeviceID )
{
  uint8_t Loop ;
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  ClearBlock ( SubDeviceID , sizeof ( DeviceIDType ) ) ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ Loop ] ;
    if ( ! Ch )
      break ;
    * SubDeviceID ++ = Ch ;
  } /* for */ ;
} /* Assign */


static uint8_t CompareFlash ( uint8_t * Text ,
                              uint16_t  Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * Text ++ ;
    Ch2 = pgm_read_byte_near ( Pattern ++ ) ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( Ch1 == 0 )
    {
      return ( true ) ;
    } /* if */ ;
  } /* for */ ;
} /* CompareFlash */


static uint8_t CompareDeviceID ( uint8_t * DeviceName ,
                                 uint8_t * Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * DeviceName ++ ;
    Ch2 = * Pattern ++ ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( ( Ch1 == null ) || ( Ch2 == null ) )
      return ( true ) ;
  } /* for */ ;
} /* CompareDeviceID */


static void CodePassword ( void )
{
  uint8_t Loop ;
  uint8_t Element ;
  uint8_t BitCorrection ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = EEBlock . Password [ Loop ] ^ 0xFF ;
    BitCorrection = Element & 0x01 ;
    Element = Element >> 1 ;
    if ( BitCorrection )
      Element |= 0x80 ;
    EEBlock . Password [ Loop ] = Element ;
  } /* for */ ;
} /* CodePassword */


static void DecodePassword ( void )
{
  uint8_t Loop ;
  uint16_t Element ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = ( uint16_t ) EEBlock . Password [ Loop ] ;
    Element = Element << 1 ;
    if ( Element >= 0x100 )
      Element ++ ;
    Element &= 0xFF ;
    Element ^= 0xFF ;
    EEBlock . Password [ Loop ] = ( uint8_t ) Element ;
  } /* for */ ;
} /* DecodePassword */


static void ReadEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  wdt_reset ( ) ;
  eeprom_read_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                      sizeof ( EEPROMBlockDataRecT ) ) ;
  DecodePassword ( ) ;
} /* ReadEEBlock */


static void WriteEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  CodePassword ( ) ;
  wdt_disable ( ) ;
  eeprom_write_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                       sizeof ( EEPROMBlockDataRecT ) ) ;
  wdt_enable ( WDTO_1S ) ;
  DecodePassword ( ) ;
} /* WriteEEBlock */


static uint16_t ComputeCRC ( uint16_t  Length ,
                             uint8_t * Buffer )
{
  uint16_t Loop ;
  uint16_t ResultCRC ;
  uint16_t WorkCRC ;
  /*----------------------------------------------------------------*/
  ResultCRC = 0xFFFF ;
  for ( Loop = 0 ; Loop < Length ; Loop ++ )
  {
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ Buffer [ Loop ] & 0x0F ] ;
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ ( Buffer [ Loop ] >> 4 ) & 0x0F ] ;
  } /* for */ ;
  return ( ResultCRC ) ;
} /* ComputeCRC */


static void WordToStr ( uint8_t * Number ,
                        uint8_t   ConvWidth ,
                        uint8_t   EmptyCh ,
                        uint16_t  Value )
{
  uint8_t Loop ;
  uint8_t LocalBuffer [ 8 ] ;
  uint8_t LastDigit ;
  /*----------------------------------------------------------------*/
  if ( ConvWidth > 7 )
  {
    * Number ++ = '!' ;
    * Number = ( uint8_t ) 0 ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < ConvWidth ; Loop ++ )
    LocalBuffer [ Loop ] = EmptyCh ;
  LocalBuffer [ ConvWidth ] = null ;
  Loop = ConvWidth - 1 ;
  for ( ; ; )
  {
    LastDigit = ( uint8_t ) ( Value % 10 ) ;
    LocalBuffer [ Loop ] = LastDigit + ( uint8_t ) 0x30 ;
    Value = Value / 10 ;
    if ( ( ! Loop ) || ( ! Value ) )
      break ;
    Loop -- ;
  } /* for */ ;
  for ( Loop = 0 ; Loop <= ConvWidth ; Loop ++ )
    * Number ++ = LocalBuffer [ Loop ] ;
} /* WordToStr */


static uint8_t * BegOfStr ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  while ( * String == Space )
    String ++ ;
  return ( String ) ;
} /* BegOfStr */


static uint8_t StringToNumb ( uint8_t * String )
{
  uint8_t Ch ;
  uint8_t RetValue ;
  /*----------------------------------------------------------------*/
  RetValue = 0 ;
  for ( ; ; )
  {
    Ch = * String ++ ;
    if ( ! Ch )
      break ;
    if ( ( Ch < '0' ) || ( Ch > '9' ) )
    {
      RetValue = 0 ;
      break ;
    } /* if */ ;
    RetValue = 10 * RetValue + Ch - 0x30 ;
  } /* for */ ;
  return ( RetValue ) ;
} /* StringToNumb */


static void EEBlockOperation ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  ReadEEBlock ( ) ;
  if ( EEBlock . EEBlockMetric != StandardSygnature )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
    return ;
  } /* if */ ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  if ( CRCValue != 0 )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
  } /* if */ ;
} /* EEBlockOperation */


static uint8_t MeDestination ( void )
{
  /*----------------------------------------------------------------*/
  return ( CompareDeviceID ( SerialService . DestinationDeviceName , EEBlock . MyDeviceID ) ) ;
} /* MeDestination */


static uint8_t GeneralDestination ( void )
{
  /*----------------------------------------------------------------*/
  return ( CompareFlash ( SerialService . DestinationDeviceName , ( uint16_t ) GeneralPoolDeviceName ) ) ;
} /* GeneralDestination */


static void AddToCyclicReplyDevicesID ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicTxt ( SerialService . SourceDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( SerialService . DestinationDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
} /* AddToCyclicReplyDevicesID */


static void AddToCyclicReplyForCommand ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclic ( 'R' ) ;
  AddToCyclicTxt ( SerialService . CommandText ) ;
  AddToCyclic ( '.' ) ;
} /* AddToCyclicReplyForCommand */


static void AddToCyclicCommandReplyOK ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) OK ) ;
} /* AddToCyclicCommandReplyOK */


static void AddToCyclicCommandReplyFail ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) Fail ) ;
} /* AddToCyclicCommandReplyFail */


static void SendKeyName ( uint8_t KeyCode )
{
  /*----------------------------------------------------------------*/
  if ( KeyCode == StandKey0Code )
  {
    AddToCyclicFlashTxt ( ( uint16_t ) StandKey0CodeName ) ;
    return ;
  } /* if */ ;
  if ( KeyCode == LongKey0Code )
  {
    AddToCyclicFlashTxt ( ( uint16_t ) LongKey0CodeName ) ;
    return ;
  } /* if */ ;
} /* SendKeyName */


static uint8_t DecodeKeyName ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  if ( CompareFlash ( String , ( uint16_t ) StandKey0CodeName ) )
    return ( StandKey0Code ) ;
  if ( CompareFlash ( String , ( uint16_t ) LongKey0CodeName ) )
    return ( LongKey0Code ) ;
  return ( 0xFF ) ;
} /* DecodeKeyName */



static void ConnectCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = EEBlock . Password [ Loop ] ;
    if ( SerialService . ParamText [ Loop ] != Ch )
      return ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  ConnectContext = true ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ConnectCommandExecute */


static void DisConnectCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ConnectContext = false ;
  AddToCyclicCommandReplyOK ( ) ;
} /* DisConnectCommandExecute */


static void SetDeviceNameCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  Assign ( EEBlock . MyDeviceID ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* SetDeviceNameCommandExecute */


static void HelloCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) DeviceName ) ;
} /* HelloCommandExecute */


static void GetVersionCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) HardwareSoftwareVersion ) ;
} /* GetVersionCommandExecute */


static void GetIDCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( cr ) ;
} /* GetIDCommandExecute */


static void ReinitCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlockOperation ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ReinitCommandExecute */


static void ResetEECommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  SetDefaultEEBlock ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* ResetEECommandExecute */


static void SaveEECommandExecute ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  EEBlock . CRCValue = 0 ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = CRCValue ;
  WriteEEBlock ( ) ;
  AddToCyclicCommandReplyOK ( ) ;
} /* SaveEECommandExecute */


static void GetChnStatusCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( ParamSeparator ) ;
  if ( CurrentChannelIsOn )
  {
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOnState ) ;
  } /* if ... */
  else
  {
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOffState ) ;
  } /* if ... else */ ;
} /* GetChnStatusCommandExecute */


static void AddSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( ! EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      Assign ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* AddSuperiorChnCommandExecute */


static void DelSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( EEBlock . EventDestinDeviceID [ Loop ] , SerialService . ParamText ) )
    {
      ClearBlock ( EEBlock . EventDestinDeviceID [ Loop ] , sizeof ( DeviceIDType ) ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* DelSuperiorChnCommandExecute */


static void ListSuperiorChnCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t EmptyTable ;
  /*----------------------------------------------------------------*/
  EmptyTable = true ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      EmptyTable = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  if ( EmptyTable )
  {
    AddToCyclicReplyDevicesID ( ) ;
    AddToCyclicReplyForCommand ( ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      AddToCyclicReplyDevicesID ( ) ;
      AddToCyclicReplyForCommand ( ) ;
      AddToCyclicTxt ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* ListSuperiorChnCommandExecute */


static void ChStateServiceOnCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( ! TempEchoDevices [ Loop ] [ 0 ] )
    {
      Assign ( TempEchoDevices [ Loop ] ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* ChStateServiceOnCommandExecute */


static void ChStateServiceOffCommandExecute ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( CompareDeviceID ( TempEchoDevices [ Loop ] , SerialService . ParamText ) )
    {
      ClearBlock ( TempEchoDevices [ Loop ] , sizeof ( DeviceIDType ) ) ;
      AddToCyclicCommandReplyOK ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* ChStateServiceOffCommandExecute */


static void ChStateServiceListCommandExecute ( void )
{
  uint8_t Loop ;
  uint8_t EmptyTable ;
  /*----------------------------------------------------------------*/
  EmptyTable = true ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      EmptyTable = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  if ( EmptyTable )
  {
    AddToCyclicReplyDevicesID ( ) ;
    AddToCyclicReplyForCommand ( ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) EmptyFlagTxt ) ;
    return ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      AddToCyclicReplyDevicesID ( ) ;
      AddToCyclicReplyForCommand ( ) ;
      AddToCyclicTxt ( TempEchoDevices [ Loop ] ) ;
      AddToCyclic ( cr ) ;
    } /* if */ ;
  } /* for */ ;
} /* ChStateServiceListCommandExecute */


static void DisplayTableElement ( PtrKeyCommDescrRecT CommDescr ,
                                  uint8_t             KeyCode )
{
  uint8_t Number [ 8 ] ;
  /*----------------------------------------------------------------*/
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  SendKeyName ( KeyCode ) ;
  AddToCyclic ( ParamSeparator ) ;
  AddToCyclicTxt ( CommDescr -> CommandText ) ;
  if ( CommDescr -> ParamValue )
  {
    AddToCyclic ( ParamSeparator ) ;
    WordToStr ( Number , 3 , Space , CommDescr -> ParamValue ) ;
    AddToCyclicTxt ( BegOfStr ( Number ) ) ;
  } /* if */ ;
  AddToCyclic ( cr ) ;
} /* DisplayTableElement */


static void ListKeyTableCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  DisplayTableElement ( & EEBlock . StandardKeySerrvice , StandKey0Code ) ;
  DisplayTableElement ( & EEBlock . LongKeyService , LongKey0Code ) ;
} /* ListKeyTableCommandExecute */


static uint8_t DecodeCommandName ( uint8_t * String )
{
  /*----------------------------------------------------------------*/
  if ( CompareFlash ( String , ( uint16_t ) ChOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChOffCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChInvCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChPulseOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChPulseOffCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChLongPulseOnCommand ) )
    return ( true ) ;
  if ( CompareFlash ( String , ( uint16_t ) ChLongPulseOffCommand ) )
    return ( true ) ;
  return ( false ) ;
} /* DecodeCommandName */


static uint8_t UnpackData ( KeyCommParamRecT * KeyCmmDescr )
{
  DeviceIDType Tmp ;
  uint8_t Loop ;
  uint8_t Ch ;
  uint8_t ArrIndex ;
  /*----------------------------------------------------------------*/
  ClearBlock ( KeyCmmDescr , sizeof ( KeyCommParamRecT ) ) ;
  ClearBlock ( Tmp , sizeof ( DeviceIDType ) ) ;
  ArrIndex = 0 ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ ArrIndex ++ ] ;
    if ( ( Ch < 0x20 ) || ( Ch == ParamSeparator ) )
    {
      break ;
    } /* if */ ;
    Tmp [ Loop ] = Ch ;
  } /* for */ ;
  KeyCmmDescr -> CommKey = DecodeKeyName ( Tmp ) ;
  if ( KeyCmmDescr -> CommKey == 0xFF )
  {
    return ( false ) ;
  } /* if */ ;
  if ( Ch != ParamSeparator )
  {
    return ( false ) ;
  } /* if */ ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = SerialService . ParamText [ ArrIndex ++ ] ;
    if ( ( Ch < 0x20 ) || ( Ch == ParamSeparator ) )
    {
      break ;
    } /* if */ ;
    KeyCmmDescr -> CommDescr . CommandText [ Loop ] = Ch ;
  } /* for */ ;
  if ( ! DecodeCommandName ( KeyCmmDescr -> CommDescr . CommandText ) )
  {
    return ( false ) ;
  } /* if */ ;
  if ( Ch == ParamSeparator )
  {
    ClearBlock ( Tmp , sizeof ( DeviceIDType ) ) ;
    for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
    {
      Ch = SerialService . ParamText [ ArrIndex ++ ] ;
      if ( Ch < 0x20 )
      {
        break ;
      } /* if */ ;
      Tmp [ Loop ] = Ch ;
    } /* for */ ;
    KeyCmmDescr -> CommDescr . ParamValue = StringToNumb ( Tmp ) ;
    if ( ! KeyCmmDescr -> CommDescr . ParamValue )
    {
      return ( false ) ;
    } /* if */ ;
  } /* if */ ;
  return ( true ) ;
} /* UnpackData */


static void AddKeyTableCommandExecute ( void )
{
  KeyCommParamRecT AddElement ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! UnpackData ( & AddElement ) )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  UnpackData ( & AddElement ) ;
  if ( AddElement . CommKey == StandKey0Code )
  {
    EEBlock . StandardKeySerrvice = AddElement . CommDescr ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  if ( AddElement . CommKey == LongKey0Code )
  {
    EEBlock . LongKeyService = AddElement . CommDescr ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* AddKeyTableCommandExecute */


static void DelKeyTableCommandExecute ( void )
{
  KeyCommParamRecT DelElement ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( ! UnpackData ( & DelElement ) )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  if ( DelElement . CommKey == StandKey0Code )
  {
    ClearBlock ( & EEBlock . StandardKeySerrvice , sizeof ( KeyCommDescrRecT ) ) ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  if ( DelElement . CommKey == LongKey0Code )
  {
    ClearBlock ( & EEBlock . LongKeyService , sizeof ( KeyCommDescrRecT ) ) ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* DelKeyTableCommandExecute */


static void NewPasswordCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  if ( SerialService . ParamText [ 0 ] )
  {
    for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
    {
      Ch = SerialService . ParamText [ Loop ] ;
      EEBlock . Password [ Loop ] = Ch ;
      if ( Ch == null )
        break ;
    } /* for */ ;
    AddToCyclicCommandReplyOK ( ) ;
    return ;
  } /* if */ ;
  AddToCyclicCommandReplyFail ( ) ;
} /* NewPasswordCommandExecute */


static void ExecuteKeyCommand ( uint8_t Ch ) ;

static void RunKeyCommandExecute ( void )
{
  uint8_t KeyCode ;
  /*----------------------------------------------------------------*/
  KeyCode = DecodeKeyName ( SerialService . ParamText ) ;
  if ( KeyCode == 0x0FF )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  ExecuteKeyCommand ( KeyCode ) ;
} /* RunKeyCommandExecute */



static void AddEventText ( uint8_t * DestDeviceID )
{
  /*----------------------------------------------------------------*/
  AddToCyclicTxt ( DestDeviceID ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( PartSeparator ) ;
  if ( CurrentChannelIsOn )
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOnState ) ;
  else
    AddToCyclicFlashTxt ( ( uint16_t ) ChannelIsOffState ) ;
} /* AddEventText */


static void PermEchoTransmit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < EventDestinDeviceArraySize ; Loop ++ )
  {
    if ( EEBlock . EventDestinDeviceID [ Loop ] [ 0 ] )
    {
      AddEventText ( EEBlock . EventDestinDeviceID [ Loop ] ) ;
    } /* if */ ;
  } /* for */ ;
} /* PermEchoTransmit */


static void TempEchoTransmit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    if ( TempEchoDevices [ Loop ] [ 0 ] )
    {
      AddEventText ( TempEchoDevices [ Loop ] ) ;
    } /* if */ ;
  } /* for */ ;
} /* TempEchoTransmit */


static void ChOnOperation ( void )
{
  /*----------------------------------------------------------------*/
  CurrentChannelIsOn = true ;
  PowerPort &= ~ ( 1 << PwrSwitchPin ) ;
  LEDPort &= ~ ( 1 << PowerOnOffLed ) ;
  PermEchoTransmit ( ) ;
  TempEchoTransmit ( ) ;
} /* ChOnOperation */


static void ChOffOperation ( void )
{
  /*----------------------------------------------------------------*/
  CurrentChannelIsOn = false ;
  PowerPort |= ( 1 << PwrSwitchPin ) ;
  LEDPort |= ( 1 << PowerOnOffLed ) ;
  PermEchoTransmit ( ) ;
  TempEchoTransmit ( ) ;
} /* ChOffOperation */


static void ChOnCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ChOnOperation ( ) ;
  ServiceDelayTime = 0 ;
  DelayedService = NULL ;
} /* ChOnCommandExecute */


static void ChOffCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ChOffOperation ( ) ;
  ServiceDelayTime = 0 ;
  DelayedService = NULL ;
} /* ChOffCommandExecute */


static void ChInvCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( CurrentChannelIsOn )
    ChOffCommandExecute ( ) ;
  else
    ChOnCommandExecute ( ) ;
} /* ChInvCommandExecute */


static void PulseCommandExecute ( OperationProcT RunNowService ,
                                  OperationProcT RunDelayedService ,
                                  uint8_t        UnitPrescaller )
{
  uint16_t ParamValue ;
  /*----------------------------------------------------------------*/
  if ( ! SerialService . ParamText [ 0 ] )
  {
    return ;
  } /* if */ ;
  ParamValue = StringToNumb ( SerialService . ParamText ) ;
  if ( ! ParamValue )
  {
    return ;
  } /* if */ ;
  RunNowService ( ) ;
  cli ( ) ;
  ServiceDelayTime = ( uint32_t ) UnitPrescaller * ( uint32_t ) ParamValue * ( uint32_t ) TimerPrescaller ;
  DelayedService = RunDelayedService ;
  sei ( ) ;
} /* PulseCommandExecute */


static void ChPulseOnCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  PulseCommandExecute ( ChOnOperation , ChOffOperation , 1 ) ;
} /* ChPulseOnCommandExecute */


static void ChPulseOffCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  PulseCommandExecute ( ChOffOperation , ChOnOperation , 1 ) ;
} /* ChPulseOffCommandExecute */


static void ChLongPulseOnCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  PulseCommandExecute ( ChOnOperation , ChOffOperation , 60 ) ;
} /* ChLongPulseOnCommandExecute */


static void ChLongPulseOffCommandExecute ( void )
{
  PulseCommandExecute ( ChOffOperation , ChOnOperation , 60 ) ;
} /* ChLongPulseOffCommandExecute */


static void ReplyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( CyclicDataPresent ( ) )
  {
    FlushCyclic ( ) ;
  } /* if ... */
  else
  {
    SendSerialTxt ( SerialService . SourceDeviceName ) ;
    SendSerial ( PartSeparator ) ;
    SendSerialTxt ( SerialService . DestinationDeviceName ) ;
    SendSerial ( PartSeparator ) ;
    SendSerialFlashTxt ( ( uint16_t ) NoDataText ) ;
  } /* if ... else */ ;
} /* ReplyCommandExecute */


static void ExecuteAssociatonCommand ( uint8_t *      String ,
                                       uint8_t        AssociationTableSize ,
                                       uint16_t       AssociationTable )
{
  uint8_t Loop ;
  CommandNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CommandService = ( CommandExecuteServiceProcT ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    AssociationElement . CommandName = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( CompareFlash ( String , ( uint16_t ) AssociationElement . CommandName ) )
    {
      AssociationElement . CommandService ( ) ;
      return ;
    } /* if */ ;
  } /* for */ ;
} /* ExecuteAssociatonCommand */


static void ExecuteCommand ( void )
{
  uint8_t FrameToMe ;
  /*----------------------------------------------------------------*/
  FrameToMe = MeDestination ( ) ;
  if ( ! FrameToMe )
  {
    FrameToMe = GeneralDestination ( ) ;
    if ( FrameToMe )
      StrCopy ( SerialService . DestinationDeviceName ,
                EEBlock . MyDeviceID ) ;
  } /* if */ ;
  if ( FrameToMe )
  {
    ExecuteAssociatonCommand ( SerialService . CommandText ,
                               CommandNameAssociationTabeSize ,
                               ( uint16_t ) CommandNameAssociationTabe ) ;
  } /* if */ ;
} /* ExecuteCommand */


static void ResetOperation ( void ) ;

static void DestinationIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . DestDeviceNameIndex < DeviceIDSize )
  {
    SerialService . DestinationDeviceName [ SerialService . DestDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* DestinationIDPutCh */


static void SourceIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . SrcDeviceNameIndex < DeviceIDSize )
  {
    SerialService . SourceDeviceName [ SerialService . SrcDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* SourceIDPutCh */


static void CommandIDPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . CommandTextIndex < CommandBufferSize )
  {
    SerialService . CommandText [ SerialService . CommandTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* CommandIDPutCh */


static void ParamPutCh ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  if ( SerialService . ParamTextIndex < ParamBufferSize )
  {
    SerialService . ParamText [ SerialService . ParamTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* ParamPutCh */


static void SelectParam ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      ParamPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    ParamPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectParam */


static void SelectCommand ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      CommandIDPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
    {
      CommandIDPutCh ( null ) ;
      SerialService . SerialService = SelectParam ;
      return ;
    } /* if */;
    CommandIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectCommand */


static void SelectSourceService ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      SourceIDPutCh ( null ) ;
      SerialService . SerialService = SelectCommand ;
      return ;
    } /* if */;
    SourceIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectSourceService */


static void SelectDestOperation ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      DestinationIDPutCh ( null ) ;
      SerialService . SerialService = SelectSourceService ;
      return ;
    } /* if */;
    DestinationIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectDestOperation */


static void KeyPulseCommandExecute ( OperationProcT RunNowService ,
                                     OperationProcT RunDelayedService ,
                                     uint16_t       ParamValue )
{
  /*----------------------------------------------------------------*/
  RunNowService ( ) ;
  cli ( ) ;
  ServiceDelayTime = ( uint32_t ) ParamValue * ( uint32_t ) TimerPrescaller ;
  DelayedService = RunDelayedService ;
  sei ( ) ;
} /* KeyPulseCommandExecute */


static void ChOnKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ChOnCommandExecute ( ) ;
} /* ChOnKeyCommandExecute */


static void ChOffKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ChOffCommandExecute ( ) ;
} /* ChOffKeyCommandExecute */


static void ChInvKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  ChInvCommandExecute ( ) ;
} /* ChInvKeyCommandExecute */


static void ChPulseOnKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  KeyPulseCommandExecute ( ChOnCommandExecute , ChOffCommandExecute ,
                           KeyCommandParam ) ;
} /* ChPulseOnKeyCommandExecute */


static void ChPulseOffKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  KeyPulseCommandExecute ( ChOffCommandExecute , ChOnCommandExecute ,
                           KeyCommandParam ) ;
} /* ChPulseOffKeyCommandExecute */


static void ChLongPulseOnKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  KeyPulseCommandExecute ( ChOnCommandExecute , ChOffCommandExecute ,
                           60 * KeyCommandParam ) ;
} /* ChLongPulseOnKeyCommandExecute */


static void ChLongPulseOffKeyCommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  KeyPulseCommandExecute ( ChOffCommandExecute , ChOnCommandExecute ,
                           60 * KeyCommandParam ) ;
} /* ChLongPulseOffKeyCommandExecute */


static void ExecuteKeyCommand ( uint8_t KeyC )
{
  uint8_t * CommandText ;
  /*----------------------------------------------------------------*/
  switch ( KeyC )
  {
    case StandKey0Code      :
      CommandText = EEBlock . StandardKeySerrvice . CommandText ;
      KeyCommandParam = EEBlock . StandardKeySerrvice . ParamValue ;
      break ;
    case LongKey0Code       :
      CommandText = EEBlock . LongKeyService . CommandText ;
      KeyCommandParam = EEBlock . LongKeyService . ParamValue ;
      break ;
    default                 :
      return ;
  } /* switch */ ;
  ExecuteAssociatonCommand ( CommandText , KeyCommandAssociationTabeSize ,
                             ( uint16_t ) KeyCommandAssociationTabe ) ;
} /* ExecuteKeyCommand */


static void ResetOperation ( void )
{
  /*----------------------------------------------------------------*/
  ClearBlock ( SerialService . CommandText , sizeof ( CommandBufferType ) ) ;
  ClearBlock ( SerialService . ParamText , sizeof ( ParamBufferType ) ) ;
  ClearBlock ( SerialService . DestinationDeviceName , sizeof ( DeviceIDType ) ) ;
  ClearBlock ( SerialService . SourceDeviceName , sizeof ( DeviceIDType ) ) ;
  SerialService . CommandTextIndex = 0 ;
  SerialService . ParamTextIndex = 0 ;
  SerialService . DestDeviceNameIndex = 0 ;
  SerialService . SrcDeviceNameIndex = 0 ;
  SerialService . SerialService = SelectDestOperation ;
} /* ResetOperation */


static void SoftwareInit ( void )
{
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  BlockEECurrentAddress = 0xFFFF ;
  ConnectContext = false ;
  ServiceDelayTime = 0 ;
  TimerCounter = 0 ;
  CyclicBuffer . CycFr = 0 ;
  CyclicBuffer . CycTo = 0 ;
  SerialService . RXDRec . CycFr = 0 ;
  SerialService . RXDRec . CycTo = 0 ;
  SerialService . TXDRec . CycFr = 0 ;
  SerialService . TXDRec . CycTo = 0 ;
  SerialService . TXDRec . State = IdleState ;
  SerialService . SerialLEDDelay = 0 ;
  for ( Loop = 0 ; Loop < TempEchoDevicesArraySize ; Loop ++ )
  {
    ClearBlock ( TempEchoDevices [ Loop ] , sizeof ( DeviceIDType ) ) ;
  } /* for */ ;
  DelayedServiceTimeEvent = 0 ;
  DelayedService = NULL ;
  ResetOperation ( ) ;
  KeybStatus = KeybIdle ;
  KeybCode = 0 ;
  KeybDelayCt = 0 ;
  KeybDelayTimer = 0 ;
  AnyKeyDelay = 0 ;
  TimeEvent = false ;
  CurrentChannelIsOn = false ;
} /* SoftwareInit */



static void LEDTest ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << PowerOnOffLed ) | ( 1 << StatusLed ) |
         ( 1 << RxTxLed ) | ( 1 << KeyLed ) ;
  LEDPort &= ~ Bits ;
  TimeEvent = 0 ;
  TimerCounter = 0 ;
  for ( ; ; )
  {
    wdt_reset ( ) ;
    if ( TimeEvent )
    {
      TimeEvent = false ;
      break ;
    } /* if */ ;
  } /* for */ ;
  LEDPort |= Bits ;
} /* LEDTest */


int main ( void )
{
  /*----------------------------------------------------------------*/
  HardwareInit ( Speed38400bps ) ;
  EnvirInit ( ) ;
  SoftwareInit ( ) ;
  USARTReceiveMode ( ) ;
  EEBlockOperation ( ) ;
  wdt_enable ( WDTO_1S ) ;
  sei ( ) ;
  LEDTest ( ) ;
  for ( ; ; )
  {
    wdt_reset ( ) ;
    if ( TimeEvent )
    {
      TimeEvent = false ;
      LEDPort ^= 1 << StatusLed ;
    } /* if */ ;
    if ( SerialService . SerialService )
      SerialService . SerialService ( ) ;
    if ( KeyPressed ( ) )
    {
      ExecuteKeyCommand ( ReadKey ( ) ) ;
    } /* if */ ;
    if ( DelayedServiceTimeEvent )
    {
      if ( DelayedService )
        DelayedService ( ) ;
      DelayedService = NULL ;
      DelayedServiceTimeEvent = false ;
    } /* if */ ;
  } /* for */ ;
  return ( 0 ) ;
} /* main */


